# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

__version__ = '1.2.4'
__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2020, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'
